# -*- coding: utf-8
# $Id: Portfolio.py 327 2005-09-06 21:40:47Z vahur $
#
# Copyright 2001 - 2005 by IVA Team and contributors
#
# This file is part of IVA.
#
# IVA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# IVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Portfolio - A place for blog and assignments.
"""

__version__ = "$Revision: 327 $"[11:-2]

import Globals
import WebtopFolder
import Blog
from WebtopItem import WebtopItem
from AccessControl import ClassSecurityInfo
from common import perm_view, perm_edit, perm_manage, perm_add_lo

class Portfolio(
    Blog.Blog,
    WebtopFolder.WebtopFolder,
    ):
    """ Portfolio holds assignments and blog """

    meta_type = "Portfolio"

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    portfolio_top = 1

    def __init__(self, parent, name):
	""" set id """
	WebtopItem.__init__(self, parent, name)
	self.id = 'portfolio'

    security.declareProtected(perm_view, 'index_html')
    def index_html(self, REQUEST=None):
	""" index_page """
	url = self.absolute_url()
	if 'subgroups' in url:
		return self.wt_index_html()
	return self.portfolio_index()

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
	""" set correct permissions """
	pass

    security.declareProtected(perm_view, 'get_icon')
    def get_icon(self):
	""" icon """
	return getattr(self.images, 'folder_gif')

    security.declareProtected(perm_view, 'get_name')
    def get_name(self):
	""" always portfolio """
	return "Portfolio"

    security.declareProtected(perm_view, 'getAssignments')
    def getAssignments(self):
	""" return a list of assignments (folder?) """
	return self.objectValues('AssignmentProxy')

    def createAssignmentProxy(self, a, c):
	""" sets AssignmentProxy object """
	proxy = AssignmentProxy(self, a, c)
	proxy.id = a
	self._setObject(a, proxy)
	return proxy.__of__(self)

Globals.InitializeClass(Portfolio)

class AssignmentProxy(
    WebtopFolder.WebtopFolder,
    ):
    """ Assignment folder where user can submit it's work. """
    
    meta_type = "AssignmentProxy"

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    def __init__(self, parent, assignment_id, course_id):
	""" Construct AssignmentProxy """
	WebtopItem.__init__(self, parent, '')
	self.__assignment_id = assignment_id
	self.__course_id = course_id

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
	""" ... """
	pass

    def index_html(self, REQUEST):
	""" index html """
	return self.hw_index_html()

    def __get_assignments(self):
	""" return assignments folder """
	return self.courses.get_child(self.__course_id).kodutood
	
    def __get_assignment(self):
	""" return assignment """
	return getattr(self.__get_assignments(), self.__assignment_id, None)

    security.declareProtected(perm_view, 'get_name')
    def get_name(self):
	""" return assignment id """
	as = self.__get_assignment()
	if not as or as.meta_type == 'AssignmentProxy':
		return "deleted assignment"
	else:
		print "as b", as
		return as.getTitle()
    
    security.declareProtected(perm_view, 'get_icon')
    def get_icon(self):
	""" icon """
	return getattr(self.images, 'folder_gif')

    security.declareProtected(perm_view, 'list_contents')
    def hw_list_contents(self, REQUEST, jnr):
	""" list contents + other webtop content if needed """
	my_list = self.list_contents_request(REQUEST)
	other_list = []
	hw = self.__get_assignment()
	# hw, hw==self, isinstance(hw, AssignmentProxy)
	if isinstance(hw, AssignmentProxy):
		return my_list
	hwID = hw.getAssignmentID()
	if hwID:
		users = hw.getReviewers(str(REQUEST.AUTHENTICATED_USER))
		for user in users:
			u_obj = getattr(self.fle_root(), 'fle_users').get_user_info(user)
			portfolio = u_obj.get_course_portfolio(jnr)
			o_hw = getattr(portfolio, hwID, None)
			if not o_hw: continue
			other_list += o_hw.list_contents_request(REQUEST)
	return other_list+my_list

Globals.InitializeClass(AssignmentProxy)
