# -*- coding: utf-8
# $Id: Statistics.py 314 2005-09-03 09:53:49Z vahur $
#
# Copyright 2001, 2002 by IVA Team and contributors
#
# This file is part of IVA.
#
# IVA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# IVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

""" Module collects statistics from all parts of IVA. """

__version__ = "$Revision: 314 $"[11:-2]

import time
import Globals
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Globals import Persistent, PersistentMapping
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import AccessControl
from common import perm_view,perm_manage

class StatisticsManager(PropertyManager, SimpleItem):
    """ store and server statistics. Storing should be through session expiring """

    security = AccessControl.ClassSecurityInfo()
    security.declareObjectPublic()
    
    meta_type = 'Statistics'

    security.declareProtected(perm_manage, 'statistics')
    stats_tab = PageTemplateFile('ui/stats.zpt',globals())
    stats_tab._owner = None

    manage_options = (
	{'label': 'Stats', 'action': 'stats_tab'},
	) + PropertyManager.manage_options+ SimpleItem.manage_options

    def __init__(self):
	self.id = 'Statistics'
	self.ug_stat = {} # not course spec. user stats
	self.uc_stat = {} # course spec. user stats
	self.c_stat = {} # course stats
	self.o_stat = {} # object views

    def abc(self):
	""" oo """
	self.ug_stat = {} # not course spec. user stats
	self.uc_stat = {} # course spec. user stats
	self.c_stat = {} # course stats
	self.o_stat = {} # object views
	
    def manage_afterAdd(self, item, container):
        """ manage after add 
	    storeAllURLs -- store ALL urls user goes, that can be very resource consuming.
	    trackSessions -- most of the statistics machinery is switched off. it will also disable showAllViews.
	    showAllViews -- if object is accessed then we write it down. if all access times are stored or only last
	"""
	self.manage_addProperty('storeAllURLs', 0, 'boolean')
	self.manage_addProperty('trackSessions', 1, 'boolean')
	self.manage_addProperty('showAllViews', 1, 'boolean')
	
    security.declareProtected(perm_manage, 'saveStats')
    def saveStats(self, data):
	print "data received from delScript:", data
	uname = data.get('uname', None)
	if uname is None:
		return 0
	if self.get_storeAllURLs():
		self.setAllURLs(uname, data.get('all_urls', {}))
	self.setViewedObjects(uname, data.get('obj_history', {}))
	self.setUserGlobalData(uname, data)
	self.setUserCourseData(uname, data.get('courses', []))
	return 1

    def index_html(self):
	""" index """
	return "no index here"

    def get_showAllViews(self):
	""" return showAllViews """
	return self.showAllViews

    def get_trackSessions(self):
	""" return trackSessions """
	return self.trackSessions

    def get_storeAllURLs(self):
	""" return storeAllURLs """
	return self.storeAllURLs

    def setAllURLs(self, uname, data):
	""" saves all urls users have gone through """
	ugstat = self.ug_stat.get(uname, {})
	all_urls = ugstat.get('all_urls', [])
	for x in data:
		all_urls.append(x)
	ugstat['all_urls'] = all_urls
	self.ug_stat[uname] = ugstat
	self._p_changed = True
	return 0

    def getAllURLs(self, uname):
	""" return all URLs """
	user = self.ug_stat.get(uname, [])
	if not user:
		return []
	return user.get('all_urls', [])
	return self.ug_stat.get(uname).get('all_urls', [])

    def setViewedObjects(self, uname, data):
	""" save viewed objects - webtop items mostly """
	save_all = self.get_showAllViews()
	for obj in data:
		o_sta = self.o_stat.get(obj[0], {})
		u_sta = o_sta.get(uname, {'times':[], 'count':0})

		if not save_all:
			u_sta['times'] = []
		u_sta['times'].append(obj[1])
		if not save_all:
			u_sta['count'] += obj[2]
		else:
			u_sta['count'] = -1

		o_sta[uname] = u_sta
		self.o_stat[obj[0]] = o_sta
	self._p_changed = True
	return 0

    def getViewedObjs(self):
	""" get """
	return self.o_stat

    def getLastVisitTime(self, REQUEST, url, uname):
	""" url - relative url 
	"""
	time = 0
	
	iva_url = self.fle_root().absolute_url(1)
	sess = REQUEST.SESSION.get(iva_url, None)
	if sess is not None:
		obj_history = sess.get('obj_history', None)
		if obj_history is not None:
			for x in obj_history:
				if x[0] == url:
					if x[1] > time:
						time = x[1]
			
	if self.o_stat.has_key(url):
		if self.o_stat[url].has_key(uname):
			if self.o_stat[url][uname]['times'][-1] > time:
				time = self.o_stat[url][uname]['times'][-1]
	return time

    def setUserGlobalData(self, uname, data):
	""" user global stats """
	ugstat = self.ug_stat.get(uname, {})
	ugstat['lastAccessGlobal'] = data.get('lastAccessGlobal')
	pages = ugstat.get('pagesOpenTotal', 0)
	ugstat['pagesOpenTotal'] = pages + data.get('pagesOpenTotal')
	sessions = ugstat.get('sessions', [])
	if not sessions:
		ugstat['sessions'] = []
	session = {
		'start':data.get('sessionStart'), 
		'end':data.get('lastAccessGlobal'),
		'pagesOpened': data.get('pagesOpenTotal'),
		'urls':[]
		}
	if self.get_storeAllURLs():
		session['urls'] = data.get('all_urls')
	sessions.append(session)
	ugstat['sessions'] = sessions
	self._p_changed = True
	return 1

    def getUserSessions(self, uname):
	""" return a list of user sessions """
	u = self.ug_stat.get(uname, [])
	if not u:
		return u
	return u.get('sessions', [])

    def setUserCourseData(self, uname, data):
	""" self.setUserCourseData(uname, data.get('courses', [])) """
	stat = self.uc_stat.get(uname, {})
	if not stat:
		self.uc_stat[uname] = {}
	for x in data.keys():
		cs = stat.get(x, {})
		for y in ('openPages','lastAccess','notesRead','postedNotes','postedArtefacts','postedArtefactAnnos','uploadedFiles','createdWikis','memosAdded','foldersOpened','linksAdded','testsSolved','numTimesEntered'):
			if y in ('lastAccess',):
				cs[y] = data[x].get(y)
			else:
				cs[y] = cs.get(y, 0)	# avoid KeyErrors
				cs[y] += data[x].get(y)
		vj = data[x].get('visited').get('visit_jamming')
		vk = data[x].get('visited').get('visit_knowledgebuilding')
		if vj > cs.get('visit_jamming',0):
			cs['visit_jamming'] = vj
		if vk > cs.get('visit_knowledgebuilding', 0):
			cs['visit_knowledgebuilding'] = vk
		#cs['openPages'] = data.get('openPages')
		#cs['lastAccess'] = data.get('lastAccess')
		#cs['notesRead'] = data.get('notesRead')
		#cs['postedNotes'] = data.get('postedNotes')
		#cs['postedArtefacts'] = data.get('postedArtefacts')
		#cs['postedArtefactAnnos'] = data.get('postedArtefactAnnos')
		#cs['uploadedFiles'] = data.get('uploadedFiles')
		#cs['createdWikis'] = data.get('createdWikis')
		#cs['memosAdded'] = data.get('memosAdded')
		#cs['foldersOpened'] = data.get('foldersOpened')
		#cs['linksAdded'] = data.get('linsAdded')
		#cs['testsSolved'] = data.get('testsSolved')

		#cs['numTimesEntered'] = "unimplemented"
		cs['addedEvents'] = "unimplemented"
		cs['addedQuizzes'] = "unimplemented"
		cs['addedAssignments'] = "unimplemented"
		stat[x] = cs

		csta = self.c_stat.get(x, {})
		if data[x].get('modified').get('jamming') > csta.get('modJamming',0):
			csta['modJamming'] = data[x].get('modified').get('jamming')
		if data[x].get('modified').get('knowledgebuilding') > csta.get('modKnowledge',0):
			csta['modKnowledge'] = data[x].get('modified').get('knowledgebuilding')
		if data[x].get('modified').get('bookshelf') > csta.get('modBookshelf',0):
			csta['modBookshelf'] = data[x].get('modified').get('bookshelf')
		
		csta['modAssignments'] = "unimplemented"
		csta['modQuizes'] = "unimplemented"
		csta['modEvents'] = "unimplemented"
		self.c_stat[x] = csta
	self.uc_stat[uname] = stat
	self._p_changed = True
	return 1

    def getObjVisitors(self, url):
	""" return visitors for object """
	return self.o_stat.get(url)

    def getUserCourseStat(self, course, stat, uname):
	""" return stat for user on given course """
	o = self.uc_stat.get(uname, None)
	if o is None:
		return o
	c = o.get(course, None)
	if c is None:
		return c
	return c.get(stat, None)

    def getCourseStat(self, course, stat):
	""" return stat for course """
	o = self.c_stat.get(course, None)
	if o is None:
		return o
	return o.get(stat, None)
	
Globals.InitializeClass(StatisticsManager)

class StatisticsRegister:
    """ statistics module that all other parts speak 
    
    global parameters to stat:
	    uname - for recognizing user
	    lastAccessGlobal
	    pagesOpenTotal
	    url - optional !? requires space in session data holding folder and is resource consuming

    course specific parameters:
	    openPages
	    modified - modified block ( dict )
	    	teadmuspaja - time.time
		meediapaja - time.time
		bookshelf - time.time
	    visited - visited block ( dict )
	        jamming
		knowledge
	    lastAccess
	    notesRead
	    postedNotes
	    postedArtefacts
	    postedArtefactAnnos
	    uploadedFiles
	    createdWikis
	    memosAdded
	    foldersOpened
	    linksAdded
	    testsSolved
	    numTimesEntered
    """

    security = AccessControl.ClassSecurityInfo()
    
    def __init__(self):
	""" init method """
	self.registerVariables()

    security.declarePublic('registerVariables')
    def registerVariables(self):
	""" default value for user_course_default variable.  """
	self.user_course_default = {'openPages':0,
			'modified':{'jamming':0, 'knowledgebuilding':0, 'bookshelf':0},
			'visited':{'visit_jamming':0, 'visit_knowledgebuilding':0},
			'lastAccess': time.time(),
			'notesRead':0,
			'postedNotes':0,
			'postedArtefacts':0,
			'postedArtefactAnnos':0,
			'uploadedFiles':0,
			'createdWikis':0,
			'memosAdded':0,
			'foldersOpened':0,
			'linksAdded':0,
			'testsSolved':0,
			'numTimesEntered':0
			}
			
    security.declareProtected(perm_view,'registerVisit')
    def registerVisit(self, REQUEST, course_id, userLocation):
	""" registers visit, page open, note reading etc. """
	uname = str(REQUEST.AUTHENTICATED_USER)
	my_path = self.fle_root().absolute_url(1)
	sd = REQUEST.SESSION.get(my_path, { 'pagesOpenTotal':0,
						'uname':uname,
						'courses':{}, 
						'lastAccessGlobal': time.time(),
						'sessionStart': time.time(),
						'all_urls': []
						}
				)
	paths = REQUEST.SESSION.get('iva_paths', [])

	sd['pagesOpenTotal'] += 1
	sd['lastAccessGlobal'] = time.time()
	store_all = getattr(self, 'Statistics').get_storeAllURLs()
	if store_all:
		# TODO:QUERY_STRING
		sd['all_urls'].append([REQUEST.get('URL0', 0), time.time()])
	course_id = int(course_id)
	if course_id and userLocation!='Organizer':
		ci = sd['courses'].get(course_id, None)
		if ci is None:
			import copy
			ci = copy.copy(self.user_course_default)
		ci['openPages'] += 1
		ci['lastAccess'] = time.time()
		sd['courses'][course_id] = ci

	REQUEST.SESSION.set(my_path, sd)
	if my_path not in paths:
		paths.append(my_path)
		REQUEST.SESSION.set('iva_paths', paths)
	return 1

    security.declareProtected(perm_view,'updateStat')
    def updateStat(self, REQUEST, field, course_id=None):
	""" updates one statistics field. course specific only. """
	sta_obj = getattr(self, 'Statistics')
	if not sta_obj.get_trackSessions():
		return 0
	if course_id is None:
		user = getattr(self.fle_root().fle_users, str(REQUEST.AUTHENTICATED_USER), None)
		#XXX: maybe we should check user here too ?
		course_id = int(user.get_jooksev_kursus())
	my_path = self.fle_root().absolute_url(1)
	base = REQUEST.SESSION.get(my_path, [])
	if not base:
		return 
	ci = base['courses'].get(course_id) 
	if ci is None:
		import copy
		ci = copy.copy(self.user_course_default)
	if field in ('jamming', 'knowledgebuilding', 'bookshelf'):
		ci['modified'][field] = time.time()
		#update time
	elif field in ('visit_jamming', 'visit_knowledgebuilding'):
		ci['visited'][field] = time.time()
	else:
		ci[field] += 1
	base['courses'][course_id] = ci
	REQUEST.SESSION.set(my_path, base)
	return 1

    def viewedObject(self, REQUEST, url):
	""" user is viewing object. url is basically an identifier, bare in mind that url might change
	"""
	sta_obj = getattr(self, 'Statistics') 
	if not sta_obj.get_trackSessions():
		return 0
	my_path = self.fle_root().absolute_url(1)
	base = REQUEST.SESSION.get(my_path, [])
	if not base: #session already expired?
		return
	store_all = sta_obj.get_showAllViews()
	hist = base.get('obj_history', [])
	count = 1
	if not store_all:
		tmp = []
		for x in hist:
			if x[0] != url:
				tmp.append([x[0], x[1], x[2]])
			else:
				count += x[2]
		hist = tmp
	hist.append([url,time.time(), count])
	base['obj_history'] = hist
	REQUEST.SESSION.set(my_path, base)
	return 1

    def getStatField(self, REQUEST, field, course_id=None):
	""" return one field from session. shoudn't pass course_id when accessing not course specific data"""
	my_path = self.fle_root().absolute_url(1)
	data = REQUEST.SESSION.get(my_path, [])
	if course_id:
		# do something
		cd = data.get(course_id, 0)
		if cd:
			return cd.get(field, -1)
		else:
			return 0
	else:
		return data.get(field, -1)

Globals.InitializeClass(StatisticsRegister)
